--
-- Hangul Automata for Supreme Commander: Forged Alliance
--
-- written by Joongi Kim (http://daybreaker.info)
-- based on the first project of CS322 Formal Languages and Automata, a course of Dept. of Computer Science in KAIST.
--

do -- restrict scope
	-- Internal table for Hangul components
	local choseongs = {0, 1, 3, 6, 7, 8, 16, 17, 18, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29}
	local joongseongs = {30, 31, 32, 33, 34, 35, 36, 37, 38, 39, 40, 41, 42, 43, 44, 45, 46, 47, 48, 49, 50}
	local jongseongs = {0, 1, 2, 3, 4, 5, 6, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 19, 20, 21, 22, 23, 25, 26, 27, 28, 29}

	--  The key map for 두벌식
	local keymap2 = {
		r= 0, R= 1, -- ㄱ, ㄲ
		s= 3, S= 3, -- ㄴ
		e= 6, E= 7, -- ㄷ, ㄸ
		f= 8, F= 8, -- ㄹ
		a= 16, A= 16, -- ㅁ
		q= 17, Q= 18, -- ㅂ, ㅃ
		t= 20, T= 21, -- ㅅ, ㅆ
		d= 22, D= 22, -- ㅇ
		w= 23, W= 24, -- ㅈ, ㅉ
		c= 25, C= 25, -- ㅊ
		z= 26, Z= 26, -- ㅋ
		x= 27, X= 27, -- ㅌ
		v= 28, V= 28, -- ㅍ
		g= 29, G= 29, -- ㅎ
		k= 30, K= 30, -- ㅏ
		i= 32, I= 32, -- ㅑ
		j= 34, J= 34, -- ㅓ
		u= 36, U= 36, -- ㅕ
		h= 38, H= 38, -- ㅗ
		y= 42, Y= 42, -- ㅛ
		n= 43, N= 43, -- ㅜ
		b= 47, B= 47, -- ㅠ
		m= 48, M= 48, -- ㅡ
		l= 50, L= 50, -- ㅣ
		o= 31, O= 33, -- ㅐ, ㅒ
		p= 35, P= 37, -- ㅔ, ㅖ
	}

	-- States of automaton
	local START = 0
	local CHOSEONG = 10
	local CHOSEONG2 = 12
	local JOONGSEONG = 30
	local JOONGSEONG2 = 32
	local JONGSEONG = 50
	local JONGSEONG2 = 52
	local FINISH = 90		-- CHOSEONG*에서 초성을 입력하거나, JOONGSEONG*태에서 모음을 입력한 경우
	local FINISH2 = 92		-- JOONGSEONG 또는 JONGSEONG* 상태에서 모음을 입력한 경우

	local state = nil
	local queue = {}

	-- Conversion function
	function conv2hangul(text)
		if text == "" then
			return ""
		end
		local pos = 1
		local chunks = {}

		-- Text split
		while true do
			local first, last = string.find(text, "//", pos)
			if first then
				table.insert(chunks, string.sub(text, pos, first - 1))
				pos = last + 1
			else
				table.insert(chunks, string.sub(text, pos))
				break
			end
		end
		text = ''

		-- Conversion
		local hangul_mode = false
		for i,v in ipairs(chunks) do
			if hangul_mode and v ~= '' then
				text = text .. to_hangul2(v)
			else
				text = text .. v
			end
			hangul_mode = not hangul_mode
		end
		return text
	end

	function to_hangul2(src)
		local output = {}
		init_queue()
		state = START
		for i = 1, string.len(src), 1 do
			local c = string.byte(src, i)
			local a = string.char(c)
			if H_is_hangul(a) then
				local t = transit(a)
				if t ~= 0 then
					table.insert(output, t)
				end
			else
				table.insert(output, flush())
				table.insert(output, c)
			end
		end
		table.insert(output, flush())
		return conv2utf8(output)
	end

	-- Automaton
	function init_queue()
		queue = {-1, -1, -1, -1, -1, -1}
		return queue
	end

	function is_empty_queue()
		local is_empty = true
		for i,v in ipairs(queue) do
			if v ~= -1 then
				is_empty = false
				break
			end
		end
		return is_empty
	end

	function transit(a)
		local output = 0

		-- Input check
		k = keymap2[a]
		if k == nil then -- 입력된 키를 내부 한글 코드(Unicode compatibility Jamo)로 변환
			flush()
			return string.byte(a)
		end
		local i = nil
		local is_vowel = (k >= 30 and k <= 50)					-- 모음이라면 true
		local is_consonant = (k >= 0 and k <= 29)				-- 자음이라면 true
		local can_be_jongseong = is_consonant and not (choose_index(k, {7, 18, 24}) ~= nil)	-- 종성이 될 수 있는지 여부
		
		-- State Transition
		if state == START then -- 오토마타의 시작
			if is_consonant then
				state = CHOSEONG
			else
				state = JOONGSEONG
			end
		elseif state == CHOSEONG then
			if is_vowel then
				state = JOONGSEONG
			elseif H_is_choseong_pair(queue[1], k) then
				state = CHOSEONG2
			else
				state = FINISH
			end
		elseif state == CHOSEONG2 then
			if is_vowel then
				state = JOONGSEONG
			else
				state = FINISH
			end
		elseif state == JOONGSEONG then
			if can_be_jongseong then
				state = JONGSEONG
			elseif H_is_joongseong_pair(queue[3], k) then
				state = JOONGSEONG2
			elseif is_vowel then
				state = FINISH2
			else
				state = FINISH
			end
		elseif state == JOONGSEONG2 then
			if can_be_jongseong then
				state = JONGSEONG
			elseif is_vowel then
				state = FINISH2
			else
				state = FINISH
			end
		elseif state == JONGSEONG then
			if is_consonant and H_is_jongseong_pair(queue[5], k) then
				state = JONGSEONG2
			elseif is_vowel then
				state = FINISH2
			else
				state = FINISH
			end
		elseif state == JONGSEONG2 then
			if is_vowel then
				state = FINISH2
			else
				state = FINISH
			end
		end

		-- 초성/중성/종성 별로 코드를 저장하고 적절한 출력 및 넘김 처리를 한다.
		i = choose(state, {CHOSEONG, CHOSEONG2, JOONGSEONG, JOONGSEONG2, JONGSEONG, JONGSEONG2}, {1, 2, 3, 4, 5, 6})
		if i ~= nil then
			queue[i] = k
		end
		if state == FINISH then

			output = conv2unicode(queue)	-- 글자 합성

			init_queue()
			queue[1] = k
			state = CHOSEONG

		elseif state == FINISH2 then

			local prev_key = queue[5]
			queue[5] = -1
			if queue[6] ~= -1 then -- 겹받침 쪼개기
				queue[5] = prev_key
				prev_key = queue[6]
				queue[6] = -1
			end

			output = conv2unicode(queue)	-- 글자 합성

			init_queue()
			queue[1] = prev_key
			queue[3] = k
			state = JOONGSEONG

		end

		return output
	end

	function flush()
		local output = 0
		if not is_empty_queue() then
			output = conv2unicode(queue)
		end
		init_queue()
		state = START
		return output
	end

	-- Automaton Helpers
	function H_is_choseong_pair(k1, k2)
		local choseong_pairs = {
			{0, 0}, {6, 6}, {17, 17}, {20, 20}, {23, 23}, -- ㄲ,ㄸ,ㅃ,ㅆ,ㅉ
		}
		for i,v in ipairs(choseong_pairs) do
			if v[1] == k1 and v[2] == k2 then
				return true
			end
		end
		return false
	end

	function H_is_joongseong_pair(k1, k2)
		local joongseong_pairs = {
			{30, 50}, {32, 50}, {34,50}, {36,50}, -- ㅐ,ㅒ,ㅔ,ㅖ
			{38, 30}, {38, 31}, {38, 50}, -- ㅘ,ㅙ,ㅚ
			{43, 34}, {43, 35}, {43, 50}, -- ㅝ,ㅞ,ㅟ
			{48, 50}, -- ㅢ
		}
		for i,v in ipairs(joongseong_pairs) do
			if v[1] == k1 and v[2] == k2 then
				return true
			end
		end
		return false
	end

	function H_is_jongseong_pair(k1, k2)
		local jongseong_pairs = {
			{0, 0}, {0, 20}, -- ㄲ,ㄳ
			{3, 23}, {3, 29}, -- ㄵ,ㄶ
			{8, 0}, {8, 16}, {8, 17}, {8, 20}, {8, 27}, {8, 28}, {8, 29}, -- ㄺ,ㄻ,ㄼ,ㄽ,ㄾ,ㄿ,ㅀ,
			{17, 20}, {20, 20}, -- ㅄ,ㅆ
		}
		for i,v in ipairs(jongseong_pairs) do
			if v[1] == k1 and v[2] == k2 then
				return true
			end
		end
		return false
	end

	function H_is_hangul(a)
		if keymap2[a] == nil then
			return false
		else
			return true
		end
	end

	function choose_index(key, array)
		for i,k in ipairs(array) do
			if key == k then
				return i
			end
		end
		return nil
	end

	function choose(key, array, array2)
		for i,k in ipairs(array) do
			if key == k then
				return array2[i]
			end
		end
		return nil
	end

	-- Unicode to UTF-8 Converter
	function conv2unicode(queue)
		local jamo_only = false
		local has_jongseong = false
		local c1, c2, c3 = -1, -1, -1
		local t = nil
	
		local q = {}
		for i = 1,6,1 do
			table.insert(q, queue[i])
		end
		
		--- 자음 처리
		--  쌍자음 연타 처리
		if q[1] == 0 and q[2] == 0 then
			q[1] = 1
			q[2] = -1
		end
		if q[1] == 6 and q[2] == 6 then
			q[1] = 7
			q[2] = -1
		end
		if q[1] == 17 and q[2] == 17 then
			q[1] = 18
			q[2] = -1
		end
		if q[1] == 20 and q[2] == 20 then
			q[1] = 21
			q[2] = -1
		end
		if q[1] == 23 and q[2] == 23 then
			q[1] = 24
			q[2] = -1
		end
		c1 = q[1]

		--- 모음 처리
		--  합성 모음 처리
		if q[3] == 38 then
			t = choose(q[4], {30, 31, 50}, {39, 40, 41})
			if t ~= nil then
				q[3] = t
				q[4] = -1
			end
		end
		if q[3] == 43 then
			t = choose(q[4], {34, 35, 50}, {44, 45, 46})
			if t ~= nil then
				q[3] = t
				q[4] = -1
			end
		end
		if q[3] == 50 then
			t = choose(q[3], {48, 30, 32, 34, 36}, {49, 31, 33, 35, 37})
			if t ~= nil then
				q[3] = t
				q[4] = -1
			end
		end

		c2 = q[3]

		--- 받침 처리
		if q[5] ~= -1 then
			has_jongseong = true
		end
		-- 합성 받침 처리
		if q[5] == 0 then
			t = choose(q[6], {0, 20}, {1, 2})
			if t ~= nil then
				q[5] = t
				q[6] = -1
			end
		end
		if q[5] == 3 then
			t = choose(q[6], {23, 29}, {4, 5})
			if t ~= nil then
				q[5] = t
				q[6] = -1
			end
		end
		if q[5] == 8 then
			t = choose(q[6], {0, 16, 17, 20, 27, 28, 29}, {9, 10, 11, 12, 13, 14, 15})
			if t ~= nil then
				q[5] = t
				q[6] = -1
			end
		end
		if q[5] == 17 and q[6] == 20 then
			q[5] = 19
			q[6] = -1
		end
		if q[5] == 20 and q[6] == 20 then
			q[5] = 21
			q[6] = -1
		end
		
		c3 = q[5]

		--- 단독 자소 처리
		if q[1] ~= -1 and q[3] == -1 then -- 자음만 있는 경우
			if has_jongseong then
				return 0
			end
			c1 = q[1]
			jamo_only = true
		end
		if q[1] == -1 and q[3] ~= -1 then -- 모음만 있는 경우
			if has_jongseong then
				return 0
			end
			c1 = q[3]
			jamo_only = true
		end

		--- 최종 코드 생성
		if jamo_only then
			return 0x3131 + c1
		else
			c1 = choose_index(c1, {0, 1, 3, 6, 7, 8, 16, 17, 18, 20, 21, 22, 23, 24, 25, 26, 27, 28, 29})
			c2 = c2 - 30
			c3 = choose(c3, {-1, 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 18, 19, 20, 21, 22, 23, 24, 25 ,26 ,27, 28, 29},
							{0, 1, 2, 3, 4, 5, 6, 7, 0, 8, 9, 10, 11, 12, 13, 14, 15, 16, 17, 0, 18, 19, 20, 21, 22, 0, 23, 24, 25, 26, 27})
			if c1 == nil or c2 == -31 or c3 == nil then
				return 0
			end
			c1 = c1 - 1
		end
		return 0xAC00 + c1 * 588 + c2 * 28 + c3

	end

	function conv2utf8(unicode_list)
		local result = ''
		local w,x,y,z = 0,0,0,0
		local function modulo(a, b)
			return a - math.floor(a/b) * b
		end
		for i,v in ipairs(unicode_list) do
			if v ~= 0 and v ~= nil then
				if v <= 0x7F then -- same as ASCII
					result = result .. string.char(v)
				elseif v >= 0x80 and v <= 0x7FF then -- 2 bytes
					--[[
					y = (v & 0x0007C0) >> 6
					z = v & 0x00003F
					]]--
					y = math.floor(modulo(v, 0x000800) / 64)
					z = modulo(v, 0x000040)
					result = result .. string.char(0xC0 + y, 0x80 + z)
				elseif (v >= 0x800 and v <= 0xD7FF) or (v >= 0xE000 and v <= 0xFFFF) then -- 3 bytes
					--[[
					x = (v & 0x00F000) >> 12
					y = (v & 0x000FC0) >> 6
					z = v & 0x00003F
					]]--
					x = math.floor(modulo(v, 0x010000) / 4096)
					y = math.floor(modulo(v, 0x001000) / 64)
					z = modulo(v, 0x000040)
					result = result .. string.char(0xE0 + x, 0x80 + y, 0x80 + z)
				elseif (v >= 0x10000 and v <= 0x10FFFF) then -- 4 bytes
					--[[
					w = (v & 0x1C0000) >> 18
					x = (v & 0x03F000) >> 12
					y = (v & 0x000FC0) >> 6
					z = v & 0x00003F
					]]--
					w = math.floor(modulo(v, 0x200000) / 262144)
					x = math.floor(modulo(v, 0x040000) / 4096)
					y = math.floor(modulo(v, 0x001000) / 64)
					z = modulo(v, 0x000040)
					result = result .. string.char(0xF0 + w, 0x80 + x, 0x80 + y, 0x80 + z)
				end
			end
		end
		return result
	end
end

-- vim: set noet ts=4 sts=4 sw=4 fenc=utf8: --
